public class Biblioteca {
	private String nome;
	private NodoLista init;

	public Biblioteca(String n) {
		nome=n;
		init=null;
	}

	/*
	 * Prende in input un Libro e ritorna true o false a seconda se questo è o meno presente in biblioteca
	 */
	private boolean presente(Libro l)
	{
		NodoLista aux = init;
		while(aux!=null && !aux.info.equals(l))
		aux = aux.next;
		if (aux==null)
			return false;
		else 
			return true;
	}	

	/*
		Prende in input un codice e ritorna il relativo Libro se esso è presente, null altrimenti
	*/
	private Libro presente(int codice)
	{
		NodoLista aux = init;
		while(aux!=null && ((Libro) aux.info).codice!=codice)
		aux = aux.next;
		if (aux==null)
			return null;
		else 
			return (Libro) aux.info;
	}	

	/*
	 * Stampa il nome ed i libri contenuti nella biblioteca
	 */
	private void stampa() {
		
		System.out.println("Biblioteca "+nome+":");
		for (int i=0;i<80;i++) System.out.print("*"); System.out.print("\n");
		System.out.println("Codice\t| Titolo\t| Autore");
		for (int i=0;i<80;i++) System.out.print("-"); System.out.print("\n");
		NodoLista aux = init;
		while(aux!=null) {
			System.out.println(aux.info);
			aux = aux.next;
		}
		for (int i=0;i<80;i++) System.out.print("*"); System.out.print("\n");
	}	

	
	/*
		quantiLibriAutore: che, dato un autore, restituisce il numero di suoi libri presenti nella biblioteca;
	*/
	public int quantiLibriAutore(String aut) {
		int counter=0;
		NodoLista aux = init;
		while(aux!=null) {
			if (((Libro) aux.info).autore.equals(aut))
				counter++;
			aux = aux.next;
		}
		return counter;
	}

	/*
	 * Precondizione: la lista "in" contiene una sequenza di stringhe
	 */
	public Lista verificaAutori(Lista in) {
		Lista out=new Lista();
		while (!in.vuota()) {
			String aut=(String) in.primo();
			if (quantiLibriAutore(aut)==0)
				out=out.inserisciPrimo(aut);
			in=in.eliminaPrimo();
		}
		return out;
	}

	/*
		tuttiLibriAutore: che, dato un autore, restituisce un array di stringhe contenente i codici di tutti i libri 
		dell'autore specificato; oppure null se non esistono libri per l'autore specificato.
		Questa implementazione sfruttando quantiLibriAutore per determinare la dimensione dell'array codici, 
		scandisce due volte la lista
	*/
	public int[] tuttiLibriAutore(String aut) {
		int[] codici=new int[quantiLibriAutore(aut)];
		
		int counter=0;
		NodoLista aux = init;
		while(aux!=null) {
			if ( ((Libro) aux.info).autore.equals(aut))
				codici[counter++]=((Libro) aux.info).codice;
			aux = aux.next;
		}
		return codici;
	}
	
	/*
		tuttiLibriAutoreOttimizzata: dato un autore, restituisce un array di stringhe contenente i codici di tutti i libri dell'autore specificato; oppure null se non esistono libri per l'autore specificato;
		A differenza della precedente implementazione in questo caso costruiamo una lista d'appoggio contenente solo i 
		codici dei libri scritti dall'autore passato in input (tipicamente più piccola rispetto alla lista originaria).
		 Per costruire l'array da ritornare scandiamo quindi solamente la lista d'appoggio.
		Si noti che la complessità asintotica dei due metodi è comunque identica e pari ad O(2N)
	*/
	public int[] tuttiLibriAutoreOttimizzata(String aut) {
		NodoLista init2=new NodoLista(null,null);
		NodoLista aux2=init2; 
		
		int counter=0;
		NodoLista aux = init;
		while(aux!=null) {
			if (((Libro) aux.info).autore.equals(aut)) {
				aux2.next=new NodoLista(new Integer(((Libro) aux.info).codice),null);
				aux2=aux2.next;
				counter++;
			}
			aux = aux.next;
		}

		if (counter==0) return null;
		
		int[] codici=new int[counter];
		init2=init2.next; // rimuove il nodo generatore

		for (int i=0; i<counter; i++) 	{
			codici[i]=((Integer) init2.info).intValue();
			init2=init2.next;
		}
		return codici;
	}

	/*
	 * Metodo ausiliare usato in tuttiAutori per verificare se nella lista di autori
	 * costruita visitando la lista dei libri è già presente un dato autore.
	 */
	private boolean autorePresente(String a, NodoLista aut) {
		while (aut!=null) {
			if (aut.info.equals(a)) return true;
			aut=aut.next;
		}
		return false;
	}
	
	/*
		tuttiAutori: che restituisce un array di stringhe contenente i nomi di tutti gli autori i cui libri sono presenti nella biblioteca;
	*/	
	public String[] tuttiAutori() {
		NodoLista init2=new NodoLista(null,null);
		NodoLista aux=init;
		int counter=0;		
		while (aux!=null) {
			String aut=((Libro) aux.info).autore;
			//verifica se l'autore è già presente nella lista d'appoggio
			if (!autorePresente(aut,init2.next)) {
				NodoLista temp=new NodoLista(aut,init2.next);
				init2.next=temp;
				counter++;
				/* Alternativamente si poteva aggiungere in coda
				aux2.next=new NodoListaAutore(aut,null);	
				aux2=aux2.next;
				*/
			}
			aux=aux.next;
		}
		if (counter==0) return null;
			
		String[] autori=new String[counter];
		init2=init2.next; // rimuove il nodo generatore
	
		for (int i=0; i<counter; i++) 	{
			autori[i]=(String) init2.info;
			init2=init2.next;
		}
		return autori;		
	}
	
	/*
		eliminaLibriAutore: che, dato un autore, elimina dalla biblioteca tutti i suoi libri;
	*/
	public void eliminaLibriAutore(String aut) throws Exception {
		NodoLista g=new NodoLista(null,init); //nodo generatore
		NodoLista cur=g.next, prec=g;

		while(cur!=null) {
			if (( (Libro) cur.info).autore.equals(aut))
				prec.next=cur.next;
			else 
				prec=cur;
			cur=cur.next;
		}
		init=g.next; // rimuove il nodo generatore
   	}

	/*
	 * aggiungi: dati il codice, l'autore e il titolo di un libro, lo aggiunge alla biblioteca; 
	 * se è già presente un libro con lo stesso codice genera una eccezione
	 */
	public void aggiungi(int Codice, String Autore, String Titolo) throws Exception
	{
		Libro l=new Libro(Codice,Autore,Titolo);
		if(presente(l))
			throw new Exception("Libro già inserito");
		else 
		{
			init = new NodoLista (l,init);
			/* Equivalente a:
			NodoLista aux = new NodoLista ();
			aux.info = l;
			aux.next = init;
			init=aux;*/
		} 
	}
	
	/*
	 * Risposta al quesito 2. Come garantire l'ordinamento della lista rispetto al titolo
	 */
	public void aggiungiOrdinato(int Codice, String Autore, String Titolo) throws Exception
	{
		Libro l=new Libro(Codice,Autore,Titolo);
		
		NodoLista g=new NodoLista(null,init); //nodo generatore
		NodoLista cur=g.next, prec=g;
		

		while(cur!=null && ( (Libro) cur.info).titolo.compareTo(l.titolo)<0) {
			prec=cur;
			cur=cur.next;
		}

		if (cur!=null && ( (Libro) cur.info).titolo.equals(l.titolo) )
			throw new Exception("Libro già inserito");
		
		prec.next=new NodoLista(l,cur);

		init=g.next; // rimuove il nodo generatore
		
	}

	/* autore: che, dato il codice di un libro, ne restituisce l'autore; se non esiste un libro con quel codice genera una eccezione; 
	*/
	public String autore(int codice) throws Exception {
		Libro l=presente(codice);
		if (l== null)
			 throw new Exception("Libro con codice "+codice+" non presente"); 
	   	return l.autore;
	}

	/* titolo: che, dato il codice di un libro, ne restituisce il titolo; se non esiste un libro con quel codice genera una eccezione; 
 	*/
	public String titolo(int codice) throws Exception {
		Libro l=presente(codice);
		if (l== null)
			 throw new Exception("Libro con codice "+codice+" non presente"); 
	   	return l.titolo;
	}

	/* elimina: che, dato il codice di un libro, lo elimina dalla biblioteca; 
	 * se il libro non è presente nella biblioteca genera una eccezione;
	 */
	public void elimina(int codice) throws Exception {
		if (init==null)
			throw new Exception("La biblioteca non contiene alcun libro"); 
		
		if (( (Libro) init.info).codice==codice) {
			// eliminazione del primo nodo della lista
			init = init.next;
		}	
		else {
			NodoLista p = init;
			while (p.next!=null && ( (Libro) p.next.info).codice!=codice)
				p = p.next;
			if (p.next!=null)
				// p e' il riferimento al nodo che precede 
				// il nodo contenente x
				p.next = p.next.next;
			else 
				throw new Exception("Libro con codice "+codice+" non presente"); 
		}
   	}


	/*
	 * totaleLibri: che restituisce il numero totale di libri nella biblioteca.
	 * NOTA: Per ottimizzare questo metodo - ovvero renderlo O(1) nel tempo -
	 * avremmo potuto utilizzare un ulteriore variabile (private) per memorizzare
	 * il numero di libri presenti nella biblioteca. Avremmo pertanto dovuto 
	 * aggiornare tale variabile in occasione
	 */
	public int totaleLibri() {
		int counter=0;
		NodoLista aux = init;
		while(aux!=null) {
			counter++;
			aux = aux.next;
		}
		return counter;
	}


	public static void main(String args[]) throws Exception {
		
		int NUM_LIBRI=10000;
		int NUM_AUTORI=NUM_LIBRI/10;
		
		Biblioteca b=new Biblioteca("TEST");
		
		b.aggiungiOrdinato(1,"Dante Alighieri","Divina Commedia");
		b.aggiungiOrdinato(2,"Boccaccio","Decameron");
		b.aggiungiOrdinato(3,"Alessandro Manzoni","I Promessi Sposi");
		b.aggiungiOrdinato(4,"Alessandro Manzoni","Poesie e Odi");
		
		b.stampa();
		
		System.out.println("lancio verificaAutori(...) con Manzoni, Alighieri, Kafka e Moravia nella lista di ingresso.");
		Lista l=new Lista();
		l=l.inserisciPrimo("Alessandro Manzoni");
		l=l.inserisciPrimo("Dante Alighieri");
		l=l.inserisciPrimo("Kafka");
		l=l.inserisciPrimo("Moravia");
	
		l=b.verificaAutori(l);
		System.out.println("Autori non presenti:");
		while (!l.vuota()) {
			System.out.println(l.primo());
			l=l.eliminaPrimo();
		}
		System.out.println("***************");
		System.out.println("Numero di libri di Alessandro Manzoni:"+b.quantiLibriAutore("Alessandro Manzoni"));
		System.out.println("***************");
		System.out.println("Codici Libri di Alessandro Manzoni (versione base):");

		long lt=System.nanoTime();
		int[] cod=b.tuttiLibriAutore("Alessandro Manzoni");
		for (int i=0; i<cod.length; i++)
			System.out.println(cod[i]);
		System.out.println(System.nanoTime() - lt+" nanosec execution time.");
		System.out.println("***************");
		System.out.println("Codici Libri di Alessandro Manzoni (versione ottimizzata):");
		lt=System.nanoTime();
		cod=b.tuttiLibriAutoreOttimizzata("Alessandro Manzoni");
		for (int i=0; i<cod.length; i++)
			System.out.println(cod[i]);
		System.out.println(System.nanoTime() - lt+" nanosec execution time.");
		System.out.println("***************");
		
		System.out.print("invoco tuttiAutori()...");
		String[] autori=b.tuttiAutori();
		System.out.println("fatto!");
		for (int i=0; i<autori.length; i++)
			System.out.println(autori[i]);
		System.out.println("***************");
	
		System.out.println("totaleLibri():"+b.totaleLibri());	
		System.out.println("***************");
	
		System.out.println("Elimino libri di Alessandro Manzoni.");
		b.eliminaLibriAutore("Alessandro Manzoni");
		System.out.println("***************");
		System.out.println("Libri di Alessandro Manzoni:"+
							b.quantiLibriAutore("Alessandro Manzoni"));
		System.out.println("***************");
		b.stampa();
	
		for (int i=1; i<=2; i++) {
			try {
			System.out.println("Autore del libro con codice "+i+": "+b.autore(i));
			}
			catch (Exception e) {e.printStackTrace();}
		}
		
		for (int i=1; i<=3; i++){
			System.out.println("Elimino libro con codice "+i);
			try {
				b.elimina(i);
			}
			catch (Exception e) {	
					b.stampa();
					e.printStackTrace();
			}
		}
		
		System.out.print("Popolo la biblioteca con "+NUM_LIBRI+" libri associati a "+NUM_AUTORI+" autori...");
		
		for (int i=0; i<NUM_LIBRI; i++)
			b.aggiungi(i,"Autore:"+i%NUM_AUTORI,"Titolo:"+i);
	
		System.out.println("fatto!");
		b.aggiungi(NUM_LIBRI+1,"Alessandro Manzoni","I Promessi Sposi");
		b.aggiungi(NUM_LIBRI+1,"Alessandro Manzoni","Poesie e Odi");
		
		System.out.println("***************");
		System.out.println("Numero di libri di Alessandro Manzoni:"+b.quantiLibriAutore("Alessandro Manzoni"));
		System.out.println("***************");
		System.out.println("Codici Libri di Alessandro Manzoni (versione base):");
				
		lt=System.nanoTime();
		cod=b.tuttiLibriAutore("Alessandro Manzoni");
		for (int i=0; i<cod.length; i++)
			System.out.println(cod[i]);
		System.out.println(System.nanoTime() - lt+" nanosec execution time.");
		System.out.println("***************");
		System.out.println("Codici Libri di Alessandro Manzoni (versione ottimizzata):");
		lt=System.nanoTime();
		cod=b.tuttiLibriAutoreOttimizzata("Alessandro Manzoni");
		for (int i=0; i<cod.length; i++)
			System.out.println(cod[i]);
		System.out.println(System.nanoTime() - lt+" nanosec execution time.");
		System.out.println("***************");
		
	}
  	
}